#!C:\bld\refnx_1562915010631\_h_env\pythonw.exe
# EASY-INSTALL-ENTRY-SCRIPT: 'refnx==0.1.8','gui_scripts','refnx'
__requires__ = 'refnx==0.1.8'
import re
import sys
from pkg_resources import load_entry_point

if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(
        load_entry_point('refnx==0.1.8', 'gui_scripts', 'refnx')()
    )
